# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common constants used for explainability."""

import tensorflow.compat.v1 as tf

DEFAULT_SERVING_SIGNATURE_DEF_KEY = (
    tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY)

# Output index for when the model's output is a scalar.
SCALAR_OUTPUT_INDEX = -1

# Metadata specifies names of tensors to be used to explain TF models.
METADATA_FILE_NAME = "explanation_metadata.json"
OVERLAY_METADATA_FILE_NAME = "explanation_overlay_metadata.json"

# Baseline and example scores in attribution outputs.
ATTR_BASELINE_SCORE = "__baseline_score__"
ATTR_EXAMPLE_SCORE = "__example_score__"

# Approximation error for each attribution method.
# The way of calculating the value for different methods can be different but
# it should give users insight on tweaking the method's parameters.
ATTR_APPROX_ERROR = "__approx_error__"

# The key in attribution dict that holds post processed image response.
B64_JPEG = "b64_jpeg"

# The maximum number of baselines users can specify in the metadata.
MAX_NUM_BASELINES = 10

# Post processor registered names for visualizing attributions.
PIXELS = "pixels"
OUTLINES = "outlines"

# Strategy names
SAMPLED_SHAPLEY = "SampledShapley"
INTEGRATED_GRADIENTS = "IntegratedGradients"
XRAI = "XRAI"

# Constant strings used in explain_server
XSERVER_ATTRIBUTIONS = "attributions"
XSERVER_PROFILING = "runtime_profiling"

# Debugging related strings for explain() call
STATS_LOGGING = "xai_stats_logging"  # used as key in debug_info
STATS_LOGGING_RESULT = "xai_stats_logging_result"  # used as key in debug_info
STATS_OBJECT = "xai_stats_object"  # used as key in thread local storage


# Stats Logging key strings:
SL_EXPLAIN_PREP_DATA = ("explain: prep step data",)
SL_EXPLAIN_PREP_EVAL_EXP_TENSORS = ("explain: evaluate explained tensors",)
SL_EXPLAIN_PREP_EVAL_OUT_RES = ("explain: evaluate outputs response",)
SL_EXPLAIN_LABEL_ATTRIBUTIONS = ("explain: compute label_attributions",)
SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS,
    "_compute_attributions_via_strategy: compute raw attributions")
SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_EXPLAINED_TENSORS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS,
    "_compute_attributions_via_strategy: prep explained tensors")
SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_LABELD = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS,
    "_compute_attributions_via_strategy: compute label attributions")

SL_SHAPLEY_PREP_SOLVER = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapley-calculate_baselines_attributions: prep solver")
SL_SHAPLEY_PREP_DATA = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapley-calculate_baselines_attributions: prep data")
SL_SHAPLEY_GEN_COUNTERFACTUALS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapley-calculate_baselines_attributions: "
    "generating counterfactuals")
SL_SHAPLEY_PRED_COUNTERFACTUALS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapley-calculate_baselines_attributions: "
    "calculate counterfactual predictions")
SL_SHAPLEY_RAW_ATTRIBUTIONS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapley-calculate_baselines_attributions: "
    "calculate raw_attributions")

SL_SHAPLEY_CC_PREP_SOLVER = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapleyCC-calculate_baselines_attributions: prep solver")
SL_SHAPLEY_CC_PREP_DATA = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapleyCC-calculate_baselines_attributions: prep data")
SL_SHAPLEY_CC_GEN_COUNTERFACTUALS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapleyCC-calculate_baselines_attributions: "
    "generating counterfactuals")
SL_SHAPLEY_CC_PRED_COUNTERFACTUALS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapleyCC-calculate_baselines_attributions: "
    "calculate counterfactual predictions")
SL_SHAPLEY_CC_RAW_ATTRIBUTIONS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "SampledShapleyCC-calculate_baselines_attributions: "
    "calculate raw_attributions")

SL_IG_PREP_DATA = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "IG-calculate_single_baseline_attributions: prep data")
SL_IG_GEN_PATHS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "IG-calculate_single_baseline_attributions: "
    "generating paths")
SL_IG_ATTRIBUTION_LABELS = (
    *SL_EXPLAIN_LABEL_ATTRIBUTIONS_VIA_STRATEGY_RAW,
    "IG-calculate_single_baseline_attributions: "
    "Compute attributions for each label_index")
SL_IG_PATH_UNROLLLED = (
    *SL_IG_ATTRIBUTION_LABELS,
    "IG-_compute_ig_attributions: "
    "Path unrolling")
SL_IG_EVAL_SMOOTH_GRAD = (
    *SL_IG_ATTRIBUTION_LABELS,
    "IG-_compute_ig_attributions: "
    "Evaluate smoothed gradient")
SL_IG_COMPUTE_AVG_ATTRIBUTIONS = (
    *SL_IG_ATTRIBUTION_LABELS,
    "IG-_compute_ig_attributions: "
    "Compute averaged attributions")
# Colormap for creating the attributions heatmap. Each colormap is a list of 256
# RGB pixel values. The lowest attribution value to be visualized maps to the
# index 0 in the colormap, whereas the highest attribution value to be
# visualized maps to index 255 in the colormap.
RED_COLORMAP = [[0.0, 0.0, 0.0], [64.248046875, 0.0, 0.0],
                [128.49609375, 0.0, 0.0], [128.994140625, 0.0, 0.0],
                [129.4921875, 0.0, 0.0], [129.990234375, 0.0, 0.0],
                [130.48828125, 0.0, 0.0], [130.986328125, 0.0, 0.0],
                [131.484375, 0.0, 0.0], [131.982421875, 0.0, 0.0],
                [132.48046875, 0.0, 0.0], [132.978515625, 0.0, 0.0],
                [133.4765625, 0.0, 0.0], [133.974609375, 0.0, 0.0],
                [134.47265625, 0.0, 0.0], [134.970703125, 0.0, 0.0],
                [135.46875, 0.0, 0.0], [135.966796875, 0.0, 0.0],
                [136.46484375, 0.0, 0.0], [136.962890625, 0.0, 0.0],
                [137.4609375, 0.0, 0.0], [137.958984375, 0.0, 0.0],
                [138.45703125, 0.0, 0.0], [138.955078125, 0.0, 0.0],
                [139.453125, 0.0, 0.0], [139.951171875, 0.0, 0.0],
                [140.44921875, 0.0, 0.0], [140.947265625, 0.0, 0.0],
                [141.4453125, 0.0, 0.0], [141.943359375, 0.0, 0.0],
                [142.44140625, 0.0, 0.0], [142.939453125, 0.0, 0.0],
                [143.4375, 0.0, 0.0], [143.935546875, 0.0, 0.0],
                [144.43359375, 0.0, 0.0], [144.931640625, 0.0, 0.0],
                [145.4296875, 0.0, 0.0], [145.927734375, 0.0, 0.0],
                [146.42578125, 0.0, 0.0], [146.923828125, 0.0, 0.0],
                [147.421875, 0.0, 0.0], [147.919921875, 0.0, 0.0],
                [148.41796875, 0.0, 0.0], [148.916015625, 0.0, 0.0],
                [149.4140625, 0.0, 0.0], [149.912109375, 0.0, 0.0],
                [150.41015625, 0.0, 0.0], [150.908203125, 0.0, 0.0],
                [151.40625, 0.0, 0.0], [151.904296875, 0.0, 0.0],
                [152.40234375, 0.0, 0.0], [152.900390625, 0.0, 0.0],
                [153.3984375, 0.0, 0.0], [153.896484375, 0.0, 0.0],
                [154.39453125, 0.0, 0.0], [154.892578125, 0.0, 0.0],
                [155.390625, 0.0, 0.0], [155.888671875, 0.0, 0.0],
                [156.38671875, 0.0, 0.0], [156.884765625, 0.0, 0.0],
                [157.3828125, 0.0, 0.0], [157.880859375, 0.0, 0.0],
                [158.37890625, 0.0, 0.0], [158.876953125, 0.0, 0.0],
                [159.375, 0.0, 0.0], [159.873046875, 0.0, 0.0],
                [160.37109375, 0.0, 0.0], [160.869140625, 0.0, 0.0],
                [161.3671875, 0.0, 0.0], [161.865234375, 0.0, 0.0],
                [162.36328125, 0.0, 0.0], [162.861328125, 0.0, 0.0],
                [163.359375, 0.0, 0.0], [163.857421875, 0.0, 0.0],
                [164.35546875, 0.0, 0.0], [164.853515625, 0.0, 0.0],
                [165.3515625, 0.0, 0.0], [165.849609375, 0.0, 0.0],
                [166.34765625, 0.0, 0.0], [166.845703125, 0.0, 0.0],
                [167.34375, 0.0, 0.0], [167.841796875, 0.0, 0.0],
                [168.33984375, 0.0, 0.0], [168.837890625, 0.0, 0.0],
                [169.3359375, 0.0, 0.0], [169.833984375, 0.0, 0.0],
                [170.33203125, 0.0, 0.0], [170.830078125, 0.0, 0.0],
                [171.328125, 0.0, 0.0], [171.826171875, 0.0, 0.0],
                [172.32421875, 0.0, 0.0], [172.822265625, 0.0, 0.0],
                [173.3203125, 0.0, 0.0], [173.818359375, 0.0, 0.0],
                [174.31640625, 0.0, 0.0], [174.814453125, 0.0, 0.0],
                [175.3125, 0.0, 0.0], [175.810546875, 0.0, 0.0],
                [176.30859375, 0.0, 0.0], [176.806640625, 0.0, 0.0],
                [177.3046875, 0.0, 0.0], [177.802734375, 0.0, 0.0],
                [178.30078125, 0.0, 0.0], [178.798828125, 0.0, 0.0],
                [179.296875, 0.0, 0.0], [179.794921875, 0.0, 0.0],
                [180.29296875, 0.0, 0.0], [180.791015625, 0.0, 0.0],
                [181.2890625, 0.0, 0.0], [181.787109375, 0.0, 0.0],
                [182.28515625, 0.0, 0.0], [182.783203125, 0.0, 0.0],
                [183.28125, 0.0, 0.0], [183.779296875, 0.0, 0.0],
                [184.27734375, 0.0, 0.0], [184.775390625, 0.0, 0.0],
                [185.2734375, 0.0, 0.0], [185.771484375, 0.0, 0.0],
                [186.26953125, 0.0, 0.0], [186.767578125, 0.0, 0.0],
                [187.265625, 0.0, 0.0], [187.763671875, 0.0, 0.0],
                [188.26171875, 0.0, 0.0], [188.759765625, 0.0, 0.0],
                [189.2578125, 0.0, 0.0], [189.755859375, 0.0, 0.0],
                [190.25390625, 0.0, 0.0], [190.751953125, 0.0, 0.0],
                [191.25, 0.0, 0.0], [191.748046875, 0.0, 0.0],
                [192.24609375, 0.0, 0.0], [192.744140625, 0.0, 0.0],
                [193.2421875, 0.0, 0.0], [193.740234375, 0.0, 0.0],
                [194.23828125, 0.0, 0.0], [194.736328125, 0.0, 0.0],
                [195.234375, 0.0, 0.0], [195.732421875, 0.0, 0.0],
                [196.23046875, 0.0, 0.0], [196.728515625, 0.0, 0.0],
                [197.2265625, 0.0, 0.0], [197.724609375, 0.0, 0.0],
                [198.22265625, 0.0, 0.0], [198.720703125, 0.0, 0.0],
                [199.21875, 0.0, 0.0], [199.716796875, 0.0, 0.0],
                [200.21484375, 0.0, 0.0], [200.712890625, 0.0, 0.0],
                [201.2109375, 0.0, 0.0], [201.708984375, 0.0, 0.0],
                [202.20703125, 0.0, 0.0], [202.705078125, 0.0, 0.0],
                [203.203125, 0.0, 0.0], [203.701171875, 0.0, 0.0],
                [204.19921875, 0.0, 0.0], [204.697265625, 0.0, 0.0],
                [205.1953125, 0.0, 0.0], [205.693359375, 0.0, 0.0],
                [206.19140625, 0.0, 0.0], [206.689453125, 0.0, 0.0],
                [207.1875, 0.0, 0.0], [207.685546875, 0.0, 0.0],
                [208.18359375, 0.0, 0.0], [208.681640625, 0.0, 0.0],
                [209.1796875, 0.0, 0.0], [209.677734375, 0.0, 0.0],
                [210.17578125, 0.0, 0.0], [210.673828125, 0.0, 0.0],
                [211.171875, 0.0, 0.0], [211.669921875, 0.0, 0.0],
                [212.16796875, 0.0, 0.0], [212.666015625, 0.0, 0.0],
                [213.1640625, 0.0, 0.0], [213.662109375, 0.0, 0.0],
                [214.16015625, 0.0, 0.0], [214.658203125, 0.0, 0.0],
                [215.15625, 0.0, 0.0], [215.654296875, 0.0, 0.0],
                [216.15234375, 0.0, 0.0], [216.650390625, 0.0, 0.0],
                [217.1484375, 0.0, 0.0], [217.646484375, 0.0, 0.0],
                [218.14453125, 0.0, 0.0], [218.642578125, 0.0, 0.0],
                [219.140625, 0.0, 0.0], [219.638671875, 0.0, 0.0],
                [220.13671875, 0.0, 0.0], [220.634765625, 0.0, 0.0],
                [221.1328125, 0.0, 0.0], [221.630859375, 0.0, 0.0],
                [222.12890625, 0.0, 0.0], [222.626953125, 0.0, 0.0],
                [223.125, 0.0, 0.0], [223.623046875, 0.0, 0.0],
                [224.12109375, 0.0, 0.0], [224.619140625, 0.0, 0.0],
                [225.1171875, 0.0, 0.0], [225.615234375, 0.0, 0.0],
                [226.11328125, 0.0, 0.0], [226.611328125, 0.0, 0.0],
                [227.109375, 0.0, 0.0], [227.607421875, 0.0, 0.0],
                [228.10546875, 0.0, 0.0], [228.603515625, 0.0, 0.0],
                [229.1015625, 0.0, 0.0], [229.599609375, 0.0, 0.0],
                [230.09765625, 0.0, 0.0], [230.595703125, 0.0, 0.0],
                [231.09375, 0.0, 0.0], [231.591796875, 0.0, 0.0],
                [232.08984375, 0.0, 0.0], [232.587890625, 0.0, 0.0],
                [233.0859375, 0.0, 0.0], [233.583984375, 0.0, 0.0],
                [234.08203125, 0.0, 0.0], [234.580078125, 0.0, 0.0],
                [235.078125, 0.0, 0.0], [235.576171875, 0.0, 0.0],
                [236.07421875, 0.0, 0.0], [236.572265625, 0.0, 0.0],
                [237.0703125, 0.0, 0.0], [237.568359375, 0.0, 0.0],
                [238.06640625, 0.0, 0.0], [238.564453125, 0.0, 0.0],
                [239.0625, 0.0, 0.0], [239.560546875, 0.0, 0.0],
                [240.05859375, 0.0, 0.0], [240.556640625, 0.0, 0.0],
                [241.0546875, 0.0, 0.0], [241.552734375, 0.0, 0.0],
                [242.05078125, 0.0, 0.0], [242.548828125, 0.0, 0.0],
                [243.046875, 0.0, 0.0], [243.544921875, 0.0, 0.0],
                [244.04296875, 0.0, 0.0], [244.541015625, 0.0, 0.0],
                [245.0390625, 0.0, 0.0], [245.537109375, 0.0, 0.0],
                [246.03515625, 0.0, 0.0], [246.533203125, 0.0, 0.0],
                [247.03125, 0.0, 0.0], [247.529296875, 0.0, 0.0],
                [248.02734375, 0.0, 0.0], [248.525390625, 0.0, 0.0],
                [249.0234375, 0.0, 0.0], [249.521484375, 0.0, 0.0],
                [250.01953125, 0.0, 0.0], [250.517578125, 0.0, 0.0],
                [251.015625, 0.0, 0.0], [251.513671875, 0.0, 0.0],
                [252.01171875, 0.0, 0.0], [252.509765625, 0.0, 0.0],
                [253.0078125, 0.0, 0.0], [253.505859375, 0.0, 0.0],
                [254.00390625, 0.0, 0.0], [255.0, 0.0, 0.0]]

GREEN_COLORMAP = [[0.0, 0.0, 0.0], [0.0, 38.952755905, 0.0],
                  [0.0, 77.90551181, 0.0], [0.0, 78.608267715, 0.0],
                  [0.0, 79.31102362, 0.0], [0.0, 80.01377952499999, 0.0],
                  [0.0, 80.71653543, 0.0], [0.0, 81.419291335, 0.0],
                  [0.0, 82.12204724, 0.0], [0.0, 82.82480315000001, 0.0],
                  [0.0, 83.52755906, 0.0], [0.0, 84.23031496499999, 0.0],
                  [0.0, 84.93307087, 0.0], [0.0, 85.635826775, 0.0],
                  [0.0, 86.33858268, 0.0], [0.0, 87.041338585, 0.0],
                  [0.0, 87.74409449, 0.0], [0.0, 88.446850395, 0.0],
                  [0.0, 89.1496063, 0.0], [0.0, 89.85236220499999, 0.0],
                  [0.0, 90.55511811, 0.0], [0.0, 91.257874015, 0.0],
                  [0.0, 91.96062992, 0.0], [0.0, 92.663385825, 0.0],
                  [0.0, 93.36614173, 0.0], [0.0, 94.068897635, 0.0],
                  [0.0, 94.77165354, 0.0], [0.0, 95.47440944499999, 0.0],
                  [0.0, 96.17716535, 0.0], [0.0, 96.87992126, 0.0],
                  [0.0, 97.58267717, 0.0], [0.0, 98.285433075, 0.0],
                  [0.0, 98.98818898, 0.0], [0.0, 99.690944885, 0.0],
                  [0.0, 100.39370079, 0.0], [0.0, 101.096456695, 0.0],
                  [0.0, 101.7992126, 0.0], [0.0, 102.50196850500001, 0.0],
                  [0.0, 103.20472441, 0.0], [0.0, 103.907480315, 0.0],
                  [0.0, 104.61023622, 0.0], [0.0, 105.312992125, 0.0],
                  [0.0, 106.01574803, 0.0], [0.0, 106.718503935, 0.0],
                  [0.0, 107.42125984, 0.0], [0.0, 108.12401574500001, 0.0],
                  [0.0, 108.82677165, 0.0], [0.0, 109.529527555, 0.0],
                  [0.0, 110.23228346, 0.0], [0.0, 110.93503937, 0.0],
                  [0.0, 111.63779528, 0.0], [0.0, 112.34055118500001, 0.0],
                  [0.0, 113.04330709, 0.0], [0.0, 113.746062995, 0.0],
                  [0.0, 114.4488189, 0.0], [0.0, 115.151574805, 0.0],
                  [0.0, 115.85433071, 0.0], [0.0, 116.557086615, 0.0],
                  [0.0, 117.25984252, 0.0], [0.0, 117.96259842500001, 0.0],
                  [0.0, 118.66535433, 0.0], [0.0, 119.368110235, 0.0],
                  [0.0, 120.07086614, 0.0], [0.0, 120.773622045, 0.0],
                  [0.0, 121.47637795, 0.0], [0.0, 122.179133855, 0.0],
                  [0.0, 122.88188976, 0.0], [0.0, 123.58464566500001, 0.0],
                  [0.0, 124.28740157, 0.0], [0.0, 124.99015748, 0.0],
                  [0.0, 125.69291339, 0.0], [0.0, 126.395669295, 0.0],
                  [0.0, 127.0984252, 0.0], [0.0, 127.80118110499998, 0.0],
                  [0.0, 128.50393701, 0.0], [0.0, 129.206692915, 0.0],
                  [0.0, 129.90944882, 0.0], [0.0, 130.612204725, 0.0],
                  [0.0, 131.31496063, 0.0], [0.0, 132.017716535, 0.0],
                  [0.0, 132.72047244, 0.0], [0.0, 133.42322834499998, 0.0],
                  [0.0, 134.12598425, 0.0], [0.0, 134.828740155, 0.0],
                  [0.0, 135.53149606, 0.0], [0.0, 136.234251965, 0.0],
                  [0.0, 136.93700787, 0.0], [0.0, 137.63976378, 0.0],
                  [0.0, 138.34251969, 0.0], [0.0, 139.045275595, 0.0],
                  [0.0, 139.7480315, 0.0], [0.0, 140.450787405, 0.0],
                  [0.0, 141.15354331, 0.0], [0.0, 141.856299215, 0.0],
                  [0.0, 142.55905512, 0.0], [0.0, 143.261811025, 0.0],
                  [0.0, 143.96456693, 0.0], [0.0, 144.667322835, 0.0],
                  [0.0, 145.37007874, 0.0], [0.0, 146.072834645, 0.0],
                  [0.0, 146.77559055, 0.0], [0.0, 147.478346455, 0.0],
                  [0.0, 148.18110236, 0.0], [0.0, 148.88385826500001, 0.0],
                  [0.0, 149.58661417, 0.0], [0.0, 150.289370075, 0.0],
                  [0.0, 150.99212598, 0.0], [0.0, 151.69488189, 0.0],
                  [0.0, 152.3976378, 0.0], [0.0, 153.100393705, 0.0],
                  [0.0, 153.80314961, 0.0], [0.0, 154.505905515, 0.0],
                  [0.0, 155.20866142, 0.0], [0.0, 155.911417325, 0.0],
                  [0.0, 156.61417323, 0.0], [0.0, 157.316929135, 0.0],
                  [0.0, 158.01968504, 0.0], [0.0, 158.72244094500002, 0.0],
                  [0.0, 159.42519685, 0.0], [0.0, 160.127952755, 0.0],
                  [0.0, 160.83070866, 0.0], [0.0, 161.533464565, 0.0],
                  [0.0, 162.23622047, 0.0], [0.0, 162.938976375, 0.0],
                  [0.0, 163.64173228, 0.0], [0.0, 164.344488185, 0.0],
                  [0.0, 165.04724409, 0.0], [0.0, 165.75, 0.0],
                  [0.0, 166.45275591, 0.0], [0.0, 167.155511815, 0.0],
                  [0.0, 167.85826772, 0.0], [0.0, 168.561023625, 0.0],
                  [0.0, 169.26377953, 0.0], [0.0, 169.966535435, 0.0],
                  [0.0, 170.66929134, 0.0], [0.0, 171.372047245, 0.0],
                  [0.0, 172.07480315, 0.0], [0.0, 172.77755905499998, 0.0],
                  [0.0, 173.48031496, 0.0], [0.0, 174.183070865, 0.0],
                  [0.0, 174.88582677, 0.0], [0.0, 175.588582675, 0.0],
                  [0.0, 176.29133858, 0.0], [0.0, 176.994094485, 0.0],
                  [0.0, 177.69685039, 0.0], [0.0, 178.399606295, 0.0],
                  [0.0, 179.1023622, 0.0], [0.0, 179.80511811, 0.0],
                  [0.0, 180.50787402, 0.0], [0.0, 181.210629925, 0.0],
                  [0.0, 181.91338583, 0.0], [0.0, 182.61614173499999, 0.0],
                  [0.0, 183.31889764, 0.0], [0.0, 184.021653545, 0.0],
                  [0.0, 184.72440945, 0.0], [0.0, 185.427165355, 0.0],
                  [0.0, 186.12992126, 0.0], [0.0, 186.832677165, 0.0],
                  [0.0, 187.53543307, 0.0], [0.0, 188.238188975, 0.0],
                  [0.0, 188.94094488, 0.0], [0.0, 189.643700785, 0.0],
                  [0.0, 190.34645669, 0.0], [0.0, 191.049212595, 0.0],
                  [0.0, 191.7519685, 0.0], [0.0, 192.454724405, 0.0],
                  [0.0, 193.15748031, 0.0], [0.0, 193.86023622, 0.0],
                  [0.0, 194.56299213, 0.0], [0.0, 195.265748035, 0.0],
                  [0.0, 195.96850394, 0.0], [0.0, 196.671259845, 0.0],
                  [0.0, 197.37401575, 0.0], [0.0, 198.07677165500002, 0.0],
                  [0.0, 198.77952756, 0.0], [0.0, 199.482283465, 0.0],
                  [0.0, 200.18503937, 0.0], [0.0, 200.887795275, 0.0],
                  [0.0, 201.59055118, 0.0], [0.0, 202.293307085, 0.0],
                  [0.0, 202.99606299, 0.0], [0.0, 203.698818895, 0.0],
                  [0.0, 204.4015748, 0.0], [0.0, 205.104330705, 0.0],
                  [0.0, 205.80708661, 0.0], [0.0, 206.50984252, 0.0],
                  [0.0, 207.21259843, 0.0], [0.0, 207.91535433500002, 0.0],
                  [0.0, 208.61811024, 0.0], [0.0, 209.320866145, 0.0],
                  [0.0, 210.02362205, 0.0], [0.0, 210.726377955, 0.0],
                  [0.0, 211.42913386, 0.0], [0.0, 212.13188976499998, 0.0],
                  [0.0, 212.83464567, 0.0], [0.0, 213.537401575, 0.0],
                  [0.0, 214.24015748, 0.0], [0.0, 214.942913385, 0.0],
                  [0.0, 215.64566929, 0.0], [0.0, 216.348425195, 0.0],
                  [0.0, 217.0511811, 0.0], [0.0, 217.753937005, 0.0],
                  [0.0, 218.45669291, 0.0], [0.0, 219.159448815, 0.0],
                  [0.0, 219.86220472, 0.0], [0.0, 220.56496063, 0.0],
                  [0.0, 221.26771654, 0.0], [0.0, 221.97047244499998, 0.0],
                  [0.0, 222.67322835, 0.0], [0.0, 223.375984255, 0.0],
                  [0.0, 224.07874016, 0.0], [0.0, 224.781496065, 0.0],
                  [0.0, 225.48425197, 0.0], [0.0, 226.187007875, 0.0],
                  [0.0, 226.88976378, 0.0], [0.0, 227.592519685, 0.0],
                  [0.0, 228.29527559, 0.0], [0.0, 228.998031495, 0.0],
                  [0.0, 229.7007874, 0.0], [0.0, 230.403543305, 0.0],
                  [0.0, 231.10629921, 0.0], [0.0, 231.809055115, 0.0],
                  [0.0, 232.51181102, 0.0], [0.0, 233.21456692499999, 0.0],
                  [0.0, 233.91732283, 0.0], [0.0, 234.62007874, 0.0],
                  [0.0, 235.32283465, 0.0], [0.0, 236.025590555, 0.0],
                  [0.0, 236.72834646, 0.0], [0.0, 237.43110236500002, 0.0],
                  [0.0, 238.13385827, 0.0], [0.0, 238.836614175, 0.0],
                  [0.0, 239.53937008, 0.0], [0.0, 240.242125985, 0.0],
                  [0.0, 240.94488189, 0.0], [0.0, 241.647637795, 0.0],
                  [0.0, 242.3503937, 0.0], [0.0, 243.053149605, 0.0],
                  [0.0, 243.75590551, 0.0], [0.0, 244.458661415, 0.0],
                  [0.0, 245.16141732, 0.0], [0.0, 245.864173225, 0.0],
                  [0.0, 246.56692913, 0.0], [0.0, 247.269685035, 0.0],
                  [0.0, 247.97244094, 0.0], [0.0, 248.67519685000002, 0.0],
                  [0.0, 249.37795276, 0.0], [0.0, 250.080708665, 0.0],
                  [0.0, 250.78346457, 0.0], [0.0, 251.486220475, 0.0],
                  [0.0, 252.18897638, 0.0], [0.0, 252.891732285, 0.0],
                  [0.0, 253.59448819, 0.0], [0.0, 254.297244095, 0.0],
                  [0.0, 255.0, 0.0], [0.0, 255.0, 0.0]]

VIRIDIS_COLORMAP = [[68, 1, 84], [68, 2, 85], [68, 3, 87], [69, 5, 88],
                    [69, 6, 90], [69, 8, 91], [70, 9, 92], [70, 11, 94],
                    [70, 12, 95], [70, 14, 97], [71, 15, 98], [71, 17, 99],
                    [71, 18, 101], [71, 20, 102], [71, 21, 103], [71, 22, 105],
                    [71, 24, 106], [72, 25, 107], [72, 26, 108], [72, 28, 110],
                    [72, 29, 111], [72, 30, 112], [72, 32, 113], [72, 33, 114],
                    [72, 34, 115], [72, 35, 116], [71, 37, 117], [71, 38, 118],
                    [71, 39, 119], [71, 40, 120], [71, 42, 121], [71, 43, 122],
                    [71, 44, 123], [70, 45, 124], [70, 47, 124], [70, 48, 125],
                    [70, 49, 126], [69, 50, 127], [69, 52, 127], [69, 53, 128],
                    [69, 54, 129], [68, 55, 129], [68, 57, 130], [67, 58, 131],
                    [67, 59, 131], [67, 60, 132], [66, 61, 132], [66, 62, 133],
                    [66, 64, 133], [65, 65, 134], [65, 66, 134], [64, 67, 135],
                    [64, 68, 135], [63, 69, 135], [63, 71, 136], [62, 72, 136],
                    [62, 73, 137], [61, 74, 137], [61, 75, 137], [61, 76, 137],
                    [60, 77, 138], [60, 78, 138], [59, 80, 138], [59, 81, 138],
                    [58, 82, 139], [58, 83, 139], [57, 84, 139], [57, 85, 139],
                    [56, 86, 139], [56, 87, 140], [55, 88, 140], [55, 89, 140],
                    [54, 90, 140], [54, 91, 140], [53, 92, 140], [53, 93, 140],
                    [52, 94, 141], [52, 95, 141], [51, 96, 141], [51, 97, 141],
                    [50, 98, 141], [50, 99, 141],
                    [49, 100, 141], [49, 101, 141], [49, 102, 141],
                    [48, 103, 141], [48, 104, 141], [47, 105, 141],
                    [47, 106, 141], [46, 107, 142], [46, 108, 142],
                    [46, 109, 142], [45, 110, 142], [45, 111, 142],
                    [44, 112, 142], [44, 113, 142], [44, 114, 142],
                    [43, 115, 142], [43, 116, 142], [42, 117, 142],
                    [42, 118, 142], [42, 119, 142], [41, 120, 142],
                    [41, 121, 142], [40, 122, 142], [40, 122, 142],
                    [40, 123, 142], [39, 124, 142], [39, 125, 142],
                    [39, 126, 142], [38, 127, 142], [38, 128, 142],
                    [38, 129, 142], [37, 130, 142], [37, 131, 141],
                    [36, 132, 141], [36, 133, 141], [36, 134, 141],
                    [35, 135, 141], [35, 136, 141], [35, 137, 141],
                    [34, 137, 141], [34, 138, 141], [34, 139, 141],
                    [33, 140, 141], [33, 141, 140], [33, 142, 140],
                    [32, 143, 140], [32, 144, 140], [32, 145, 140],
                    [31, 146, 140], [31, 147, 139], [31, 148, 139],
                    [31, 149, 139], [31, 150, 139], [30, 151, 138],
                    [30, 152, 138], [30, 153, 138], [30, 153, 138],
                    [30, 154, 137], [30, 155, 137], [30, 156, 137],
                    [30, 157, 136], [30, 158, 136], [30, 159, 136],
                    [30, 160, 135], [31, 161, 135], [31, 162, 134],
                    [31, 163, 134], [32, 164, 133], [32, 165, 133],
                    [33, 166, 133], [33, 167, 132], [34, 167, 132],
                    [35, 168, 131], [35, 169, 130], [36, 170, 130],
                    [37, 171, 129], [38, 172, 129], [39, 173, 128],
                    [40, 174, 127], [41, 175, 127], [42, 176, 126],
                    [43, 177, 125], [44, 177, 125], [46, 178, 124],
                    [47, 179, 123], [48, 180, 122], [50, 181, 122],
                    [51, 182, 121], [53, 183, 120], [54, 184, 119],
                    [56, 185, 118], [57, 185, 118], [59, 186, 117],
                    [61, 187, 116], [62, 188, 115], [64, 189, 114],
                    [66, 190, 113], [68, 190, 112], [69, 191, 111],
                    [71, 192, 110], [73, 193, 109], [75, 194, 108],
                    [77, 194, 107], [79, 195, 105], [81, 196, 104],
                    [83, 197, 103], [85, 198, 102], [87, 198, 101],
                    [89, 199, 100], [91, 200, 98], [94, 201, 97], [96, 201, 96],
                    [98, 202, 95], [100, 203, 93], [103, 204,
                                                    92], [105, 204, 91],
                    [107, 205, 89], [109, 206, 88], [112, 206, 86],
                    [114, 207, 85], [116, 208, 84], [119, 208, 82],
                    [121, 209, 81], [124, 210, 79], [126, 210, 78],
                    [129, 211, 76], [131, 211, 75], [134, 212, 73],
                    [136, 213, 71], [139, 213, 70], [141, 214, 68],
                    [144, 214, 67], [146, 215, 65], [149, 215, 63],
                    [151, 216, 62], [154, 216, 60], [157, 217, 58],
                    [159, 217, 56], [162, 218, 55], [165, 218, 53],
                    [167, 219, 51], [170, 219, 50], [173, 220, 48],
                    [175, 220, 46], [178, 221, 44], [181, 221, 43],
                    [183, 221, 41], [186, 222, 39], [189, 222, 38],
                    [191, 223, 36], [194, 223, 34], [197, 223, 33],
                    [199, 224, 31], [202, 224, 30], [205, 224, 29],
                    [207, 225, 28], [210, 225, 27], [212, 225, 26],
                    [215, 226, 25], [218, 226, 24], [220, 226, 24],
                    [223, 227, 24], [225, 227, 24], [228, 227, 24],
                    [231, 228, 25], [233, 228, 25], [236, 228, 26],
                    [238, 229, 27], [241, 229, 28], [243, 229, 30],
                    [246, 230, 31], [248, 230, 33], [250, 230, 34],
                    [253, 231, 36]]

PINK_COLORMAP = [
    [RED_COLORMAP[i][0], 0, RED_COLORMAP[i][0]] for i in range(0, 256)
]

# Corresponds to the PiYG palette.
PINK_WHITE_GREEN_COLORMAP = [[255, i * 2, 255] for i in range(0, 128)]
PINK_WHITE_GREEN_COLORMAP.extend(
    [[255 - 2 * i, 255, 255 - 2 * i] for i in range(0, 128)])

COLOR_MAPS = {
    "red_green": {
        "positive": GREEN_COLORMAP,
        "negative": RED_COLORMAP
    },
    "red": {
        "positive": RED_COLORMAP,
        "negative": RED_COLORMAP
    },
    "green": {
        "positive": GREEN_COLORMAP,
        "negative": GREEN_COLORMAP
    },
    "viridis": {
        "positive": VIRIDIS_COLORMAP,
        "negative": VIRIDIS_COLORMAP
    },
    "pink_white_green": {
        "positive": PINK_WHITE_GREEN_COLORMAP,
        "negative": PINK_WHITE_GREEN_COLORMAP
    },
    "pink_green": {
        "positive": GREEN_COLORMAP,
        "negative": PINK_COLORMAP
    }
}
